__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from reveal_user_annotation.common.config_package import get_package_path
from reveal_user_annotation.common.datarw import get_file_row_generator


def get_topic_set(file_path):
    """
    Opens one of the topic set resource files and returns a set of topics.

    - Input:  - file_path: The path pointing to the topic set resource file.

    - Output: - topic_set: A python set of strings.
    """
    topic_set = set()
    file_row_gen = get_file_row_generator(file_path, ",")  # The separator here is irrelevant.
    for file_row in file_row_gen:
        topic_set.add(file_row[0])

    return topic_set


def get_story_set():
    """
    Returns a set of all the story-related topics found in the SNOW dataset.
    """
    file_path = get_package_path() + "/twitter/res/topics/story_set.txt"
    topics = get_topic_set(file_path)

    return topics


def get_theme_set():
    """
    Returns a set of all the thematic-related keywords found in the SNOW dataset.
    """
    file_path = get_package_path() + "/twitter/res/topics/theme_set.txt"
    topics = get_topic_set(file_path)

    return topics


def get_attribute_set():
    """
    Returns a set of all the user-related attributes (e.g. occupation) found in the SNOW dataset.
    """
    file_path = get_package_path() + "/twitter/res/topics/attribute_set.txt"
    topics = get_topic_set(file_path)

    return topics


def get_stance_set():
    """
    Returns a set of all the user-related stances (e.g. political affiliations, religious view) found in the SNOW dataset.
    """
    file_path = get_package_path() + "/twitter/res/topics/stance_set.txt"
    topics = get_topic_set(file_path)

    return topics


def get_geographical_set():
    """
    Returns a set of all the geographical attributes (e.g. uk, sweden) found in the SNOW dataset.
    """
    file_path = get_package_path() + "/twitter/res/topics/geographical_set.txt"
    topics = get_topic_set(file_path)

    return topics


def get_sport_set():
    """
    Returns a set of all the sport-related attributes (e.g. football, golf) found in the SNOW dataset.
    """
    file_path = get_package_path() + "/twitter/res/topics/sport_set.txt"
    topics = get_topic_set(file_path)

    return topics


def get_reveal_set():
    """
    Returns a set of all the topics that are interesting for REVEAL use-cases.
    """
    file_path = get_package_path() + "/twitter/res/topics/story_set.txt"
    story_topics = get_topic_set(file_path)

    file_path = get_package_path() + "/twitter/res/topics/theme_set.txt"
    theme_topics = get_topic_set(file_path)

    file_path = get_package_path() + "/twitter/res/topics/attribute_set.txt"
    attribute_topics = get_topic_set(file_path)

    file_path = get_package_path() + "/twitter/res/topics/stance_set.txt"
    stance_topics = get_topic_set(file_path)

    file_path = get_package_path() + "/twitter/res/topics/geographical_set.txt"
    geographical_topics = get_topic_set(file_path)

    topics = story_topics | theme_topics | attribute_topics | stance_topics | geographical_topics

    return topics


def get_topic_keyword_dictionary():
    """
    Opens the topic-keyword map resource file and returns the corresponding python dictionary.

    - Input:  - file_path: The path pointing to the topic-keyword map resource file.

    - Output: - topic_set: A topic to keyword python dictionary.
    """
    topic_keyword_dictionary = dict()
    file_row_gen = get_file_row_generator(get_package_path() + "/twitter/res/topics/topic_keyword_mapping" + ".txt",
                                          ",",
                                          "utf-8")
    for file_row in file_row_gen:
        topic_keyword_dictionary[file_row[0]] = set([keyword for keyword in file_row[1:]])

    return topic_keyword_dictionary



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import time
import twython
from urllib.error import URLError
from http.client import BadStatusLine

from reveal_user_annotation.text.map_data import chunks
from reveal_user_annotation.twitter.twitter_util import login, safe_twitter_request_handler


def check_suspension(user_twitter_id_list):
    """
    Looks up a list of user ids and checks whether they are currently suspended.

    Input: - user_twitter_id_list: A python list of Twitter user ids in integer format to be looked-up.

    Outputs: - suspended_user_twitter_id_list: A python list of suspended Twitter user ids in integer format.
             - non_suspended_user_twitter_id_list: A python list of non suspended Twitter user ids in integer format.
             - unknown_status_user_twitter_id_list: A python list of unknown status Twitter user ids in integer format.
    """
    ####################################################################################################################
    # Log into my application.
    ####################################################################################################################
    twitter = login()

    ####################################################################################################################
    # Lookup users
    ####################################################################################################################
    # Initialize look-up lists
    suspended_user_twitter_id_list = list()
    non_suspended_user_twitter_id_list = list()
    unknown_status_user_twitter_id_list = list()

    append_suspended_twitter_user = suspended_user_twitter_id_list.append
    append_non_suspended_twitter_user = non_suspended_user_twitter_id_list.append
    extend_unknown_status_twitter_user = unknown_status_user_twitter_id_list.extend

    # Split twitter user id list into sub-lists of length 100 (This is the Twitter API function limit).
    user_lookup_counter = 0
    user_lookup_time_window_start = time.perf_counter()
    for hundred_length_sub_list in chunks(list(user_twitter_id_list), 100):
        # Make safe twitter request.
        try:
            api_result, user_lookup_counter, user_lookup_time_window_start\
                = safe_twitter_request_handler(twitter_api_func=twitter.lookup_user,
                                               call_rate_limit=60,
                                               call_counter=user_lookup_counter,
                                               time_window_start=user_lookup_time_window_start,
                                               max_retries=10,
                                               wait_period=2,
                                               parameters=hundred_length_sub_list)
            # If the call is succesful, turn hundred sub-list to a set for faster search.
            hundred_length_sub_list = set(hundred_length_sub_list)

            # Check who is suspended and who is not.
            for hydrated_user_object in api_result:
                hydrated_twitter_user_id = hydrated_user_object["id"]

                if hydrated_twitter_user_id in hundred_length_sub_list:
                    append_non_suspended_twitter_user(hydrated_twitter_user_id)
                else:
                    append_suspended_twitter_user(hydrated_twitter_user_id)
        except twython.TwythonError:
            # If the call is unsuccesful, we do not know about the status of the users.
            extend_unknown_status_twitter_user(hundred_length_sub_list)
        except URLError:
            # If the call is unsuccesful, we do not know about the status of the users.
            extend_unknown_status_twitter_user(hundred_length_sub_list)
        except BadStatusLine:
            # If the call is unsuccesful, we do not know about the status of the users.
            extend_unknown_status_twitter_user(hundred_length_sub_list)

    return suspended_user_twitter_id_list, non_suspended_user_twitter_id_list, unknown_status_user_twitter_id_list



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import pymongo
import json
import os


def extract_snow_tweets_from_file_generator(json_file_path):
    """
    A generator that opens a file containing many json tweets and yields all the tweets contained inside.

    Input:  - json_file_path: The path of a json file containing a tweet in each line.

    Yields: - tweet: A tweet in python dictionary (json) format.
    """
    with open(json_file_path, "r", encoding="utf-8") as fp:
        for file_line in fp:
            tweet = json.loads(file_line)
            yield tweet


def extract_all_snow_tweets_from_disk_generator(json_folder_path):
    """
    A generator that returns all SNOW tweets stored in disk.

    Input:  - json_file_path: The path of the folder containing the raw data.

    Yields: - tweet: A tweet in python dictionary (json) format.
    """
    # Get a generator with all file paths in the folder
    json_file_path_generator = (json_folder_path + "/" + name for name in os.listdir(json_folder_path))

    for path in json_file_path_generator:
        for tweet in extract_snow_tweets_from_file_generator(path):
            yield tweet


def store_snow_tweets_from_disk_to_mongodb(snow_tweets_folder):
    """
    Store all SNOW tweets in a mongodb collection.
    """
    client = pymongo.MongoClient("localhost", 27017)

    db = client["snow_tweet_storage"]
    collection = db["tweets"]

    for tweet in extract_all_snow_tweets_from_disk_generator(snow_tweets_folder):
        collection.insert(tweet)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from itertools import islice, zip_longest
import numpy as np


def grouper(iterable, n, pad_value=None):
    """
    Returns a generator of n-length chunks of an input iterable, with appropriate padding at the end.

    Example: grouper(3, 'abcdefg', 'x') --> ('a','b','c'), ('d','e','f'), ('g','x','x')

    Inputs: - iterable: The source iterable that needs to be chunkified.
            - n: The size of the chunks.
            - pad_value: The value with which the last chunk will be padded.

    Output: - chunk_gen: A generator of n-length chunks of an input iterable.
    """
    chunk_gen = (chunk for chunk in zip_longest(*[iter(iterable)]*n, fillvalue=pad_value))
    return chunk_gen


def chunks(iterable, n):
    """
    A python generator that yields 100-length sub-list chunks.

    Input:  - full_list: The input list that is to be separated in chunks of 100.
            - chunk_size: Should be set to 100, unless the Twitter API changes.

    Yields: - sub_list: List chunks of length 100.
    """
    for i in np.arange(0, len(iterable), n):
        yield iterable[i:i+n]


def split_every(iterable, n):  # TODO: Remove this, or make it return a generator.
    """
    A generator of n-length chunks of an input iterable
    """
    i = iter(iterable)
    piece = list(islice(i, n))
    while piece:
        yield piece
        piece = list(islice(i, n))



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import os
import re
import string
import nltk
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords
from nltk.stem.porter import PorterStemmer
from nltk.stem.snowball import SnowballStemmer
from nltk.stem.wordnet import WordNetLemmatizer
from multiprocessing import Pool
from functools import partial
from collections import defaultdict

from reveal_user_annotation.common.config_package import get_package_path, get_threads_number
from reveal_user_annotation.common.datarw import get_file_row_generator
from reveal_user_annotation.text.map_data import chunks
from reveal_user_annotation.text.text_util import reduce_list_of_bags_of_words, combine_word_list


def clean_single_word(word, lemmatizing="wordnet"):
    """
    Performs stemming or lemmatizing on a single word.

    If we are to search for a word in a clean bag-of-words, we need to search it after the same kind of preprocessing.

    Inputs: - word: A string containing the source word.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Output: - lemma: The resulting clean lemma or stem.
    """
    if lemmatizing == "porter":
        porter = PorterStemmer()
        lemma = porter.stem(word)
    elif lemmatizing == "snowball":
        snowball = SnowballStemmer('english')
        lemma = snowball.stem(word)
    elif lemmatizing == "wordnet":
        wordnet = WordNetLemmatizer()
        lemma = wordnet.lemmatize(word)
    else:
        print("Invalid lemmatizer argument.")
        raise RuntimeError

    return lemma


def clean_document(document, lemmatizing="wordnet"):
    """
    Extracts a clean bag-of-words from a document.

    Inputs: - document: A string containing some text.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Output: - lemma_list: A python list of lemmas or stems.
            - lemma_to_keywordbag: A python dictionary that maps stems/lemmas to original topic keywords.
    """
    ####################################################################################################################
    # Tokenizing text
    ####################################################################################################################
    try:
        tokenized_document = word_tokenize(document)
    except LookupError:
        print("Warning: Could not tokenize document. If these warnings are commonplace, there is a problem with the nltk resources.")
        lemma_list = list()
        lemma_to_keywordbag = defaultdict(lambda: defaultdict(int))
        return lemma_list, lemma_to_keywordbag

    ####################################################################################################################
    # Separate ["camelCase"] into ["camel", "case"] and make every letter lower case
    ####################################################################################################################
    tokenized_document = [separate_camel_case(token).lower() for token in tokenized_document]

    ####################################################################################################################
    # Parts of speech tagger
    ####################################################################################################################
    tokenized_document = nltk.pos_tag(tokenized_document)
    tokenized_document = [token[0] for token in tokenized_document if (token[1] == "JJ" or token[1] == "NN" or token[1] == "NNS" or token[1] == "NNP")]

    ####################################################################################################################
    # Removing digits, punctuation and whitespace
    ####################################################################################################################
    # See documentation here: http://docs.python.org/2/library/string.html
    regex = re.compile('[%s]' % re.escape(string.digits + string.punctuation + string.whitespace))

    tokenized_document_no_punctuation = list()
    append_token = tokenized_document_no_punctuation.append
    for token in tokenized_document:
        new_token = regex.sub(u'', token)
        if not new_token == u'':
            append_token(new_token)

    ####################################################################################################################
    # Removing stopwords
    ####################################################################################################################
    stopset = set(stopwords.words('english'))  # Make set for faster access

    more_stopword_files_list = os.listdir(get_package_path() + "/text/res/stopwords/")
    more_stopword_files_list = (get_package_path() + "/text/res/stopwords/" + file_name for file_name in more_stopword_files_list)

    # Read more stopwords from files
    extended_stopset = list()
    append_stopwords = extended_stopset.append
    for stop_word_file in more_stopword_files_list:
        file_row_gen = get_file_row_generator(stop_word_file, ",", encoding="utf-8")
        for row in file_row_gen:
            append_stopwords(row[0])
        stopset.update(extended_stopset)

    tokenized_document_no_stopwords = list()
    append_word = tokenized_document_no_stopwords.append
    for word in tokenized_document_no_punctuation:
        if word not in stopset:
            append_word(word)

    ####################################################################################################################
    # Remove words that have been created by automated list tools.
    ####################################################################################################################
    # # TODO: This should be done either for list keywords, or with a regex test(0-9), descr(0-9).
    # tokenized_document_no_stopwords_no_autowords = list()
    # append_word = tokenized_document_no_stopwords_no_autowords.append
    # for word in tokenized_document_no_stopwords:
    #     if not word.startswith(prefix=autoword_tuple):
    #         append_word(word)

    ####################################################################################################################
    # Stemming and Lemmatizing
    ####################################################################################################################
    lemma_to_keywordbag = defaultdict(lambda: defaultdict(int))

    final_doc = list()
    append_lemma = final_doc.append
    for word in tokenized_document_no_stopwords:
        if lemmatizing == "porter":
            porter = PorterStemmer()
            stem = porter.stem(word)
            append_lemma(stem)
            lemma_to_keywordbag[stem][word] += 1
        elif lemmatizing == "snowball":
            snowball = SnowballStemmer('english')
            stem = snowball.stem(word)
            append_lemma(stem)
            lemma_to_keywordbag[stem][word] += 1
        elif lemmatizing == "wordnet":
            wordnet = WordNetLemmatizer()
            lemma = wordnet.lemmatize(word)
            append_lemma(lemma)
            lemma_to_keywordbag[lemma][word] += 1
        else:
            print("Invalid lemmatizer argument.")
            raise RuntimeError

    ####################################################################################################################
    # One more stopword removal
    ####################################################################################################################
    lemma_list = list()
    append_word = lemma_list.append
    for word in final_doc:
        if word not in stopset:
            append_word(word)

    return lemma_list, lemma_to_keywordbag


def clean_corpus_serial(corpus, lemmatizing="wordnet"):
    """
    Extracts a bag-of-words from each document in a corpus serially.

    Inputs: - corpus: A python list of python strings. Each string is a document.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Output: - list_of_bags_of_words: A list of python dictionaries representing bags-of-words.
            - lemma_to_keywordbag_total: Aggregated python dictionary that maps stems/lemmas to original topic keywords.
    """
    list_of_bags_of_words = list()
    append_bag_of_words = list_of_bags_of_words.append

    lemma_to_keywordbag_total = defaultdict(lambda: defaultdict(int))

    for document in corpus:
        word_list, lemma_to_keywordbag = clean_document(document=document, lemmatizing=lemmatizing)
        bag_of_words = combine_word_list(word_list)
        append_bag_of_words(bag_of_words)

        for lemma, keywordbag in lemma_to_keywordbag.items():
            for keyword, multiplicity in keywordbag.items():
                lemma_to_keywordbag_total[lemma][keyword] += multiplicity

    return list_of_bags_of_words, lemma_to_keywordbag_total


def extract_bag_of_words_from_corpus_parallel(corpus, lemmatizing="wordnet"):
    """
    This extracts one bag-of-words from a list of strings. The documents are mapped to parallel processes.

    Inputs: - corpus: A list of strings.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Output: - bag_of_words: This is a bag-of-words in python dictionary format.
            - lemma_to_keywordbag_total: Aggregated python dictionary that maps stems/lemmas to original topic keywords.
    """
    ####################################################################################################################
    # Map and reduce document cleaning.
    ####################################################################################################################
    # Build a pool of processes.
    pool = Pool(processes=get_threads_number()*2,)

    # Partition the tweets to chunks.
    partitioned_corpus = chunks(corpus, len(corpus) / get_threads_number())

    # Map the cleaning of the tweet corpus to a pool of processes.
    list_of_bags_of_words, list_of_lemma_to_keywordset_maps = pool.map(partial(clean_corpus_serial, lemmatizing=lemmatizing), partitioned_corpus)

    # Reduce dictionaries to a single dictionary serially.
    bag_of_words = reduce_list_of_bags_of_words(list_of_bags_of_words)

    # Reduce lemma to keyword maps to a single dictionary.
    lemma_to_keywordbag_total = defaultdict(lambda: defaultdict(int))
    for lemma_to_keywordbag in list_of_lemma_to_keywordset_maps:
        for lemma, keywordbag in lemma_to_keywordbag.items():
            for keyword, multiplicity in keywordbag.items():
                lemma_to_keywordbag_total[lemma][keyword] += multiplicity

    return bag_of_words, lemma_to_keywordbag_total


first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')
autoword_tuple = ("test", "descr")


def separate_camel_case(word):
    """
    What it says on the tin.

    Input:  - word: A string that may be in camelCase.

    Output: - separated_word: A list of strings with camel case separated.
    """
    s1 = first_cap_re.sub(r'\1 \2', word)
    separated_word = all_cap_re.sub(r'\1 \2', s1)
    return separated_word



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from collections import defaultdict

from reveal_user_annotation.text.clean_text import clean_document
from reveal_user_annotation.text.text_util import reduce_list_of_bags_of_words


def clean_twitter_list(twitter_list, lemmatizing="wordnet"):
    """
    Extracts the *set* of keywords found in a Twitter list (name + description).

    Inputs: - twitter_list: A Twitter list in json format.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Output: - keyword_set: A set of keywords (i.e. not a bag-of-words) in python set format.
            - lemma_to_keywordbag: A python dictionary that maps stems/lemmas to original topic keywords.
    """
    name_lemmas, name_lemma_to_keywordbag = clean_document(twitter_list["name"].replace("_", " ").replace("-", " "),
                                                           lemmatizing=lemmatizing)
    description_lemmas, description_lemma_to_keywordbag = clean_document(twitter_list["description"].replace("_", " ").replace("-", " "),
                                                                         lemmatizing=lemmatizing)

    keyword_set = set(name_lemmas + description_lemmas)

    lemma_to_keywordbag = defaultdict(lambda: defaultdict(int))
    for lemma, keywordbag in name_lemma_to_keywordbag.items():
        for keyword, multiplicity in keywordbag.items():
            lemma_to_keywordbag[lemma][keyword] += multiplicity

    for lemma, keywordbag in description_lemma_to_keywordbag.items():
        for keyword, multiplicity in keywordbag.items():
            lemma_to_keywordbag[lemma][keyword] += multiplicity

    return keyword_set, lemma_to_keywordbag


def clean_list_of_twitter_list(list_of_twitter_lists, lemmatizing="wordnet"):
    """
    Extracts the sets of keywords for each Twitter list.

    Inputs: - list_of_twitter_lists: A python list of Twitter lists in json format.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Output: - list_of_keyword_sets: A list of sets of keywords (i.e. not a bag-of-words) in python set format.
            - list_of_lemma_to_keywordbags: List of python dicts that map stems/lemmas to original topic keywords.
    """
    list_of_keyword_sets = list()
    append_keyword_set = list_of_keyword_sets.append

    list_of_lemma_to_keywordbags = list()
    append_lemma_to_keywordbag = list_of_lemma_to_keywordbags.append

    if list_of_twitter_lists is not None:
        for twitter_list in list_of_twitter_lists:
            if twitter_list is not None:
                keyword_set, lemma_to_keywordbag = clean_twitter_list(twitter_list=twitter_list, lemmatizing=lemmatizing)
                append_keyword_set(keyword_set)
                append_lemma_to_keywordbag(lemma_to_keywordbag)

    return list_of_keyword_sets, list_of_lemma_to_keywordbags


def user_twitter_list_bag_of_words(twitter_list_corpus, lemmatizing="wordnet"):
    """
    Extract a bag-of-words for a corpus of Twitter lists pertaining to a Twitter user.

    Inputs: - twitter_list_corpus: A python list of Twitter lists in json format.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Output: - bag_of_words: A bag-of-words in python dictionary format.
            - lemma_to_keywordbag_total: Aggregated python dictionary that maps stems/lemmas to original topic keywords.
    """
    # Extract a bag-of-words from a list of Twitter lists.
    # May result in empty sets
    list_of_keyword_sets, list_of_lemma_to_keywordbags = clean_list_of_twitter_list(twitter_list_corpus, lemmatizing)

    # Reduce keyword sets.
    bag_of_words = reduce_list_of_bags_of_words(list_of_keyword_sets)

    # Reduce lemma to keywordbag maps.
    lemma_to_keywordbag_total = defaultdict(lambda: defaultdict(int))
    for lemma_to_keywordbag in list_of_lemma_to_keywordbags:
        for lemma, keywordbag in lemma_to_keywordbag.items():
            for keyword, multiplicity in keywordbag.items():
                lemma_to_keywordbag_total[lemma][keyword] += multiplicity

    return bag_of_words, lemma_to_keywordbag_total



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import pymongo


def establish_mongo_connection(mongo_uri):
    """
    What it says on the tin.

    Inputs: - mongo_uri: A MongoDB URI.

    Output: - A MongoDB client.
    """
    client = pymongo.MongoClient(mongo_uri)
    return client


def delete_database(host_name, port_name, database_name):
    """
    Delete a mongo database using pymongo. Mongo daemon assumed to be running.

    Inputs: - host_name: The host name where the mongo database is located.
            - port_name: The port through which the mongo databse is exposed.
            - database_name: The mongo database name as a python string.
    """
    client = pymongo.MongoClient(host_name, port_name)

    client.drop_database(database_name)


def delete_collection(host_name, port_name, database_name, collection_name):
    """
    Delete a mongo document collection using pymongo. Mongo daemon assumed to be running.

    Inputs: - host_name: The host name where the mongo database is located.
            - port_name: The port through which the mongo databse is exposed.
            - database_name: The mongo database name as a python string.
            - collection_name: The mongo collection as a python string.
    """
    client = pymongo.MongoClient(host_name, port_name)

    db = client[database_name]

    db.drop_collection(collection_name)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import nltk


def combine_word_list(word_list):
    """
    Combine word list into a bag-of-words.

    Input:  - word_list: This is a python list of strings.

    Output: - bag_of_words: This is the corresponding multi-set or bag-of-words, in the form of a python dictionary.
    """
    bag_of_words = dict()

    for word in word_list:
        if word in bag_of_words.keys():
            bag_of_words[word] += 1
        else:
            bag_of_words[word] = 1

    return bag_of_words


def reduce_list_of_bags_of_words(list_of_keyword_sets):
    """
    Reduces a number of keyword sets to a bag-of-words.

    Input:  - list_of_keyword_sets: This is a python list of sets of strings.

    Output: - bag_of_words: This is the corresponding multi-set or bag-of-words, in the form of a python dictionary.
    """
    bag_of_words = dict()
    get_bag_of_words_keys = bag_of_words.keys
    for keyword_set in list_of_keyword_sets:
        for keyword in keyword_set:
            if keyword in get_bag_of_words_keys():
                bag_of_words[keyword] += 1
            else:
                bag_of_words[keyword] = 1

    return bag_of_words


def augmented_tf_idf(attribute_matrix):
    """
    Performs augmented TF-IDF normalization on a bag-of-words vector representation of data.

    Augmented TF-IDF introduced in: Manning, C. D., Raghavan, P., & Schütze, H. (2008).
                                    Introduction to information retrieval (Vol. 1, p. 6).
                                    Cambridge: Cambridge university press.

    Input:  - attribute_matrix: A bag-of-words vector representation in SciPy sparse matrix format.

    Output: - attribute_matrix: The same matrix after augmented tf-idf normalization.
    """
    number_of_documents = attribute_matrix.shape[0]

    max_term_frequencies = np.ones(number_of_documents, dtype=np.float64)
    idf_array = np.ones(attribute_matrix.shape[1], dtype=np.float64)

    # Calculate inverse document frequency
    attribute_matrix = attribute_matrix.tocsc()
    for j in range(attribute_matrix.shape[1]):
        document_frequency = attribute_matrix.getcol(j).data.size
        if document_frequency > 1:
            idf_array[j] = np.log(number_of_documents/document_frequency)

    # Calculate maximum term frequencies for a user
    attribute_matrix = attribute_matrix.tocsr()
    for i in range(attribute_matrix.shape[0]):
        max_term_frequency = attribute_matrix.getrow(i).data
        if max_term_frequency.size > 0:
            max_term_frequency = max_term_frequency.max()
            if max_term_frequency > 0.0:
                max_term_frequencies[i] = max_term_frequency

    # Do augmented tf-idf normalization
    attribute_matrix = attribute_matrix.tocoo()
    attribute_matrix.data = 0.5 + np.divide(0.5*attribute_matrix.data, np.multiply((max_term_frequencies[attribute_matrix.row]), (idf_array[attribute_matrix.col])))
    attribute_matrix = attribute_matrix.tocsr()

    return attribute_matrix


def query_list_of_words(target_word, list_of_words, edit_distance=1):
    """
    Checks whether a target word is within editing distance of any one in a set of keywords.

    Inputs: - target_word: A string containing the word we want to search in a list.
            - list_of_words: A python list of words.
            - edit_distance: For larger words, we also check for similar words based on edit_distance.

    Outputs: - new_list_of_words: This is the input list of words minus any found keywords.
             - found_list_of_words: This is the list of words that are within edit distance of the target word.
    """
    # Initialize lists
    new_list_of_words = list()
    found_list_of_words = list()

    append_left_keyword = new_list_of_words.append
    append_found_keyword = found_list_of_words.append

    # Iterate over the list of words
    for word in list_of_words:
        if len(word) > 6:
            effective_edit_distance = edit_distance
        else:
            effective_edit_distance = 0  # No edit distance for small words.
        if abs(len(word)-len(target_word)) <= effective_edit_distance:
            if nltk.edit_distance(word, target_word) <= effective_edit_distance:
                append_found_keyword(word)
            else:
                append_left_keyword(word)
        else:
            append_left_keyword(word)

    return new_list_of_words, found_list_of_words


def simple_word_query(target_word, list_of_words, edit_distance=1):
    found_list_of_words = list()
    append_found_keyword = found_list_of_words.append

    for word in list_of_words:
        if len(word) > 6:
            effective_edit_distance = edit_distance
        else:
            effective_edit_distance = 0  # No edit distance for small words.
        if abs(len(word)-len(target_word)) <= effective_edit_distance:
            if nltk.edit_distance(word, target_word) <= effective_edit_distance:
                append_found_keyword(word)
            else:
                pass
        else:
            pass

    return found_list_of_words



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import scipy.sparse as spsp
import networkx as nx
import networkx.algorithms.components as nxalgcom
from pymongo import ASCENDING

from reveal_user_annotation.text.clean_text import clean_document, combine_word_list


def get_user_to_bag_of_words_dictionary(user_twitter_id_list, database):
    """
    Returns a python dictionary that maps Twitter user ids to a bag-of-words.

    Inputs: - user_twitter_id_list: A python list of Twitter user ids.
            - database: A Mongo database object.

    Output: - user_to_bag_of_words_dictionary: A python dictionary that maps Twitter user ids to a bag-of-words.
    """
    user_to_bag_of_words_dictionary = dict()
    for user_twitter_id, bag_of_words in read_bag_of_words_for_each_user(user_twitter_id_list, database):
        user_to_bag_of_words_dictionary[user_twitter_id] = bag_of_words

    return user_to_bag_of_words_dictionary


def read_bag_of_words_for_each_user(user_twitter_id_list, database):
    """
    For each Twitter user ids, it reads the corresponding preprocessed bag-of-words.

    Inputs:  - user_twitter_id_list: A python list of Twitter user ids.
             - database: A Mongo database object.

    Yields: - user_twitter_id: A Twitter user id.
            - bag_of_words: A python dictionary that maps keywords to multiplicity.
    """
    for user_twitter_id in user_twitter_id_list:
        collection_name = str(user_twitter_id)
        user_keywords_collection = database[collection_name]

        user_keywords_cursor = user_keywords_collection.find()
        bag_of_words = next(user_keywords_cursor)

        yield user_twitter_id, bag_of_words


def store_user_documents(user_document_gen, client, mongo_database_name, mongo_collection_name):
    """
    Stores Twitter list objects that a Twitter user is a member of in different mongo collections.

    Inputs: - user_document_gen: A python generator that yields a Twitter user id and an associated document list.
            - client: A pymongo MongoClient object.
            - mongo_database_name: The name of a Mongo database as a string.
            - mongo_collection_name: The name of the mongo collection as a string.
    """
    mongo_database = client[mongo_database_name]
    mongo_collection = mongo_database[mongo_collection_name]

    # Iterate over all users to be annotated and store the Twitter lists in mongo.
    for user_twitter_id, user_document_list in user_document_gen:
        document = user_document_list
        document["_id"] = int(user_twitter_id)
        mongo_collection.update({"_id": user_twitter_id}, document, upsert=True)


def read_user_documents_for_single_user_generator(user_twitter_id, mongo_database):
    """
    Stores Twitter list objects that a Twitter user is a member of in different mongo collections.

    Inputs: - user_twitter_id: A Twitter user id.
            - mongo_database: A mongo database.

    Yields: - twitter_list: A tuple containing: * A Twitter user id.
                                                * A python list containing Twitter lists in dictionary (json) format.
    """
    collection_name = str(user_twitter_id)
    collection = mongo_database[collection_name]
    cursor = collection.find()

    for twitter_list in cursor:
        yield twitter_list


def read_user_documents_generator(user_twitter_id_list, client, mongo_database_name, mongo_collection_name):
    """
    Stores Twitter list objects that a Twitter user is a member of in different mongo collections.

    Inputs: - user_twitter_id_list: A python list of Twitter user ids.
            - client: A pymongo MongoClient object.
            - mongo_database_name: The name of a Mongo database as a string.
            - mongo_collection_name: The name of the mongo collection as a string.

    Yields: - user_twitter_id: A Twitter user id.
            - twitter_list_gen: A python generator that yields Twitter lists in dictionary (json) format.
    """
    mongo_database = client[mongo_database_name]
    mongo_collection = mongo_database[mongo_collection_name]

    cursor = mongo_collection.find({"_id": {"$in": [int(user_twitter_id) for user_twitter_id in user_twitter_id_list]}})

    for documents in cursor:
        yield documents["_id"], documents


def get_collection_documents_generator(client, database_name, collection_name, spec, latest_n, sort_key):
    """
    This is a python generator that yields tweets stored in a mongodb collection.

    Tweet "created_at" field is assumed to have been stored in the format supported by MongoDB.

    Inputs: - client: A pymongo MongoClient object.
            - database_name: The name of a Mongo database as a string.
            - collection_name: The name of the tweet collection as a string.
            - spec: A python dictionary that defines higher query arguments.
            - latest_n: The number of latest results we require from the mongo document collection.
            - sort_key: A field name according to which we will sort in ascending order.

    Yields: - document: A document in python dictionary (json) format.
    """
    mongo_database = client[database_name]
    collection = mongo_database[collection_name]
    collection.create_index(sort_key)

    if latest_n is not None:
        skip_n = collection.count() - latest_n
        if collection.count() - latest_n < 0:
            skip_n = 0
        cursor = collection.find(filter=spec).sort([(sort_key, ASCENDING), ])
        cursor = cursor[skip_n:]
    else:
        cursor = collection.find(filter=spec).sort([(sort_key, ASCENDING), ])

    for document in cursor:
        yield document


def extract_graphs_and_lemmas_from_tweets(tweet_generator):
    """
    Given a tweet python generator, we encode the information into mention and retweet graphs and a lemma matrix.

    We assume that the tweets are given in increasing timestamp.

    Inputs:  - tweet_generator: A python generator of tweets in python dictionary (json) format.

    Outputs: - mention_graph: The mention graph as a SciPy sparse matrix.
             - retweet_graph: The retweet graph as a SciPy sparse matrix.
             - user_lemma_matrix: The user lemma vector representation matrix as a SciPy sparse matrix.
             - tweet_id_set: A python set containing the Twitter ids for all the dataset tweets.
             - user_id_set: A python set containing the Twitter ids for all the dataset users.
             - lemma_to_attribute: A map from lemmas to numbers in python dictionary format.
    """
    ####################################################################################################################
    # Prepare for iterating over tweets.
    ####################################################################################################################
    # These are initialized as lists for incremental extension.
    tweet_id_set = set()
    user_id_set = list()

    add_tweet_id = tweet_id_set.add
    append_user_id = user_id_set.append

    # Initialize sparse matrix arrays.
    mention_graph_row = list()
    mention_graph_col = list()

    retweet_graph_row = list()
    retweet_graph_col = list()

    user_lemma_matrix_row = list()
    user_lemma_matrix_col = list()
    user_lemma_matrix_data = list()

    append_mention_graph_row = mention_graph_row.append
    append_mention_graph_col = mention_graph_col.append

    append_retweet_graph_row = retweet_graph_row.append
    append_retweet_graph_col = retweet_graph_col.append

    append_user_lemma_matrix_row = user_lemma_matrix_row.append
    append_user_lemma_matrix_col = user_lemma_matrix_col.append
    append_user_lemma_matrix_data = user_lemma_matrix_data.append

    # Initialize dictionaries.
    id_to_node = dict()
    id_to_name = dict()
    lemma_to_attribute = dict()

    ####################################################################################################################
    # Iterate over tweets.
    ####################################################################################################################
    counter = 0
    for tweet in tweet_generator:
        # Increment tweet counter.
        counter += 1
        print(counter)

        # Extract base tweet's values.
        tweet_id = tweet["id"]
        user_id = tweet["user"]["id"]
        user_screen_name = tweet["user"]["screen_name"]

        # Map users to distinct integer numbers.
        graph_size = len(id_to_node)
        source_node = id_to_node.setdefault(user_id, graph_size)

        # Update sets, lists and dictionaries.
        add_tweet_id(tweet_id)
        id_to_name[user_id] = user_screen_name
        append_user_id(user_id)

        ################################################################################################################
        # We are dealing with an original tweet.
        ################################################################################################################
        if "retweeted_status" not in tweet.keys():
            ############################################################################################################
            # Update user-lemma frequency matrix.
            ############################################################################################################
            # Extract lemmas from the text.
            tweet_lemmas, lemma_to_keywordbag = clean_document(tweet["text"])
            bag_of_lemmas = combine_word_list(tweet_lemmas)

            # Update the user-lemma frequency matrix one-by-one.
            for lemma, multiplicity in bag_of_lemmas.items():
                # Map lemmas to distinct integer numbers.
                vocabulary_size = len(lemma_to_attribute)
                attribute = lemma_to_attribute.setdefault(lemma, vocabulary_size)

                # Add values to the sparse matrix arrays.
                append_user_lemma_matrix_row(source_node)
                append_user_lemma_matrix_col(attribute)
                append_user_lemma_matrix_data(multiplicity)

            ############################################################################################################
            # Update mention matrix.
            ############################################################################################################
            # Get mentioned user ids.
            mentioned_user_id_set = list()
            in_reply_to_user_id = tweet["in_reply_to_user_id"]
            if in_reply_to_user_id is not None:
                mentioned_user_id_set.append(in_reply_to_user_id)

                id_to_name[in_reply_to_user_id] = tweet["in_reply_to_screen_name"]
            for user_mention in tweet["entities"]["user_mentions"]:
                mentioned_user_id = user_mention["id"]
                mentioned_user_id_set.append(mentioned_user_id)

                id_to_name[mentioned_user_id] = user_mention["screen_name"]

            # We remove duplicates.
            mentioned_user_id_set = set(mentioned_user_id_set)

            # Update the mention graph one-by-one.
            for mentioned_user_id in mentioned_user_id_set:
                # Map users to distinct integer numbers.
                graph_size = len(id_to_node)
                mention_target_node = id_to_node.setdefault(mentioned_user_id, graph_size)

                append_user_id(mentioned_user_id)

                # Add values to the sparse matrix arrays.
                append_mention_graph_row(source_node)
                append_mention_graph_col(mention_target_node)

        ################################################################################################################
        # We are dealing with a retweet.
        ################################################################################################################
        else:
            # Extract base tweet's values.
            original_tweet = tweet["retweeted_status"]
            original_tweet_id = original_tweet["id"]
            original_tweet_user_id = original_tweet["user"]["id"]
            original_tweet_user_name = original_tweet["user"]["screen_name"]

            # Map users to distinct integer numbers.
            graph_size = len(id_to_node)
            original_tweet_node = id_to_node.setdefault(original_tweet_user_id, graph_size)

            # Update retweet graph.
            append_retweet_graph_row(source_node)
            append_retweet_graph_col(original_tweet_node)

            # Extract lemmas from the text.
            tweet_lemmas, lemma_to_keywordbag = clean_document(original_tweet["text"])
            bag_of_lemmas = combine_word_list(tweet_lemmas)

            # Update the user-lemma frequency matrix one-by-one.
            for lemma, multiplicity in bag_of_lemmas.items():
                # Map lemmas to distinct integer numbers.
                vocabulary_size = len(lemma_to_attribute)
                attribute = lemma_to_attribute.setdefault(lemma, vocabulary_size)

                # Add values to the sparse matrix arrays.
                append_user_lemma_matrix_row(source_node)
                append_user_lemma_matrix_col(attribute)
                append_user_lemma_matrix_data(multiplicity)

            # Get mentioned user ids.
            mentioned_user_id_set = list()
            in_reply_to_user_id = original_tweet["in_reply_to_user_id"]
            if in_reply_to_user_id is not None:
                mentioned_user_id_set.append(in_reply_to_user_id)

                id_to_name[in_reply_to_user_id] = original_tweet["in_reply_to_screen_name"]
            for user_mention in original_tweet["entities"]["user_mentions"]:
                mentioned_user_id = user_mention["id"]
                mentioned_user_id_set.append(mentioned_user_id)

                id_to_name[mentioned_user_id] = user_mention["screen_name"]

            # We remove duplicates.
            mentioned_user_id_set = set(mentioned_user_id_set)

            # Update the mention graph one-by-one.
            for mentioned_user_id in mentioned_user_id_set:
                # Map users to distinct integer numbers.
                graph_size = len(id_to_node)
                mention_target_node = id_to_node.setdefault(mentioned_user_id, graph_size)

                append_user_id(mentioned_user_id)

                # Add values to the sparse matrix arrays.
                append_mention_graph_row(source_node)
                append_mention_graph_col(mention_target_node)

            # This is the first time we deal with this tweet.
            if original_tweet_id not in tweet_id_set:
                # Update sets, lists and dictionaries.
                add_tweet_id(original_tweet_id)
                id_to_name[original_tweet_user_id] = original_tweet_user_name
                append_user_id(original_tweet_user_id)

                ########################################################################################################
                # Update user-lemma frequency matrix.
                ########################################################################################################
                # Update the user-lemma frequency matrix one-by-one.
                for lemma, multiplicity in bag_of_lemmas.items():
                    # Map lemmas to distinct integer numbers.
                    vocabulary_size = len(lemma_to_attribute)
                    attribute = lemma_to_attribute.setdefault(lemma, vocabulary_size)

                    # Add values to the sparse matrix arrays.
                    append_user_lemma_matrix_row(original_tweet_node)
                    append_user_lemma_matrix_col(attribute)
                    append_user_lemma_matrix_data(multiplicity)

                ########################################################################################################
                # Update mention matrix.
                ########################################################################################################
                # Update the mention graph one-by-one.
                for mentioned_user_id in mentioned_user_id_set:
                    # Map users to distinct integer numbers.
                    graph_size = len(id_to_node)
                    mention_target_node = id_to_node.setdefault(mentioned_user_id, graph_size)

                    append_user_id(mentioned_user_id)

                    # Add values to the sparse matrix arrays.
                    append_mention_graph_row(original_tweet_node)
                    append_mention_graph_col(mention_target_node)
            else:
                pass

    ####################################################################################################################
    # Final steps of preprocessing tweets.
    ####################################################################################################################
    # Discard any duplicates.
    user_id_set = set(user_id_set)
    number_of_users = len(user_id_set)
    # min_number_of_users = max(user_id_set) + 1

    # Form mention graph adjacency matrix.
    mention_graph_row = np.array(mention_graph_row, dtype=np.int64)
    mention_graph_col = np.array(mention_graph_col, dtype=np.int64)
    mention_graph_data = np.ones_like(mention_graph_row, dtype=np.float64)

    mention_graph = spsp.coo_matrix((mention_graph_data, (mention_graph_row, mention_graph_col)),
                                    shape=(number_of_users, number_of_users))
    mention_graph = spsp.coo_matrix(spsp.csr_matrix(mention_graph))

    # Form retweet graph adjacency matrix.
    retweet_graph_row = np.array(retweet_graph_row, dtype=np.int64)
    retweet_graph_col = np.array(retweet_graph_col, dtype=np.int64)
    retweet_graph_data = np.ones_like(retweet_graph_row, dtype=np.float64)

    retweet_graph = spsp.coo_matrix((retweet_graph_data, (retweet_graph_row, retweet_graph_col)),
                                    shape=(number_of_users, number_of_users))
    retweet_graph = spsp.coo_matrix(spsp.csr_matrix(retweet_graph))

    # Form user-lemma matrix.
    number_of_lemmas = len(lemma_to_attribute)

    user_lemma_matrix_row = np.array(user_lemma_matrix_row, dtype=np.int64)
    user_lemma_matrix_col = np.array(user_lemma_matrix_col, dtype=np.int64)
    user_lemma_matrix_data = np.array(user_lemma_matrix_data, dtype=np.float64)

    user_lemma_matrix = spsp.coo_matrix((user_lemma_matrix_data, (user_lemma_matrix_row, user_lemma_matrix_col)),
                                        shape=(number_of_users, number_of_lemmas))
    user_lemma_matrix = spsp.coo_matrix(spsp.csr_matrix(user_lemma_matrix))

    node_to_id = dict(zip(id_to_node.values(), id_to_node.keys()))

    return mention_graph, retweet_graph, user_lemma_matrix, tweet_id_set, user_id_set, node_to_id, lemma_to_attribute, id_to_name


def extract_graphs_from_tweets(tweet_generator):
    """
    Given a tweet python generator, we encode the information into mention and retweet graphs.

    We assume that the tweets are given in increasing timestamp.

    Inputs:  - tweet_generator: A python generator of tweets in python dictionary (json) format.

    Outputs: - mention_graph: The mention graph as a SciPy sparse matrix.
             - user_id_set: A python set containing the Twitter ids for all the dataset users.
             - node_to_id: A python dictionary that maps from node anonymized ids, to twitter user ids.
    """
    ####################################################################################################################
    # Prepare for iterating over tweets.
    ####################################################################################################################
    # These are initialized as lists for incremental extension.
    tweet_id_set = set()
    user_id_set = list()
    twitter_to_reveal_user_id = dict()

    add_tweet_id = tweet_id_set.add
    append_user_id = user_id_set.append

    # Initialize sparse matrix arrays.
    mention_graph_row = list()
    mention_graph_col = list()

    retweet_graph_row = list()
    retweet_graph_col = list()

    append_mention_graph_row = mention_graph_row.append
    append_mention_graph_col = mention_graph_col.append

    append_retweet_graph_row = retweet_graph_row.append
    append_retweet_graph_col = retweet_graph_col.append

    # Initialize dictionaries.
    id_to_node = dict()
    id_to_name = dict()

    ####################################################################################################################
    # Iterate over tweets.
    ####################################################################################################################
    counter = 0
    for tweet in tweet_generator:
        # Increment tweet counter.
        counter += 1
        # print(counter)

        # Extract base tweet's values.
        tweet_id = tweet["id"]
        user_id = tweet["user"]["id"]
        user_screen_name = tweet["user"]["screen_name"]

        # Map users to distinct integer numbers.
        graph_size = len(id_to_node)
        source_node = id_to_node.setdefault(user_id, graph_size)

        # Update sets, lists and dictionaries.
        add_tweet_id(tweet_id)
        id_to_name[user_id] = user_screen_name
        append_user_id(user_id)

        # twitter_to_user_id

        ################################################################################################################
        # We are dealing with an original tweet.
        ################################################################################################################
        if "retweeted_status" not in tweet.keys():
            ############################################################################################################
            # Update mention matrix.
            ############################################################################################################
            # Get mentioned user ids.
            mentioned_user_id_set = list()
            in_reply_to_user_id = tweet["in_reply_to_user_id"]
            if in_reply_to_user_id is not None:
                mentioned_user_id_set.append(in_reply_to_user_id)

                id_to_name[in_reply_to_user_id] = tweet["in_reply_to_screen_name"]
            for user_mention in tweet["entities"]["user_mentions"]:
                mentioned_user_id = user_mention["id"]
                mentioned_user_id_set.append(mentioned_user_id)

                id_to_name[mentioned_user_id] = user_mention["screen_name"]

            # We remove duplicates.
            mentioned_user_id_set = set(mentioned_user_id_set)

            # Update the mention graph one-by-one.
            for mentioned_user_id in mentioned_user_id_set:
                # Map users to distinct integer numbers.
                graph_size = len(id_to_node)
                mention_target_node = id_to_node.setdefault(mentioned_user_id, graph_size)

                append_user_id(mentioned_user_id)

                # Add values to the sparse matrix arrays.
                append_mention_graph_row(source_node)
                append_mention_graph_col(mention_target_node)

        ################################################################################################################
        # We are dealing with a retweet.
        ################################################################################################################
        else:
            # Extract base tweet's values.
            original_tweet = tweet["retweeted_status"]
            original_tweet_id = original_tweet["id"]
            original_tweet_user_id = original_tweet["user"]["id"]
            original_tweet_user_name = original_tweet["user"]["screen_name"]

            # Map users to distinct integer numbers.
            graph_size = len(id_to_node)
            original_tweet_node = id_to_node.setdefault(original_tweet_user_id, graph_size)

            # Update retweet graph.
            append_retweet_graph_row(source_node)
            append_retweet_graph_col(original_tweet_node)

            # Get mentioned user ids.
            mentioned_user_id_set = list()
            in_reply_to_user_id = original_tweet["in_reply_to_user_id"]
            if in_reply_to_user_id is not None:
                mentioned_user_id_set.append(in_reply_to_user_id)

                id_to_name[in_reply_to_user_id] = original_tweet["in_reply_to_screen_name"]
            for user_mention in original_tweet["entities"]["user_mentions"]:
                mentioned_user_id = user_mention["id"]
                mentioned_user_id_set.append(mentioned_user_id)

                id_to_name[mentioned_user_id] = user_mention["screen_name"]

            # We remove duplicates.
            mentioned_user_id_set = set(mentioned_user_id_set)

            # Update the mention graph one-by-one.
            for mentioned_user_id in mentioned_user_id_set:
                # Map users to distinct integer numbers.
                graph_size = len(id_to_node)
                mention_target_node = id_to_node.setdefault(mentioned_user_id, graph_size)

                append_user_id(mentioned_user_id)

                # Add values to the sparse matrix arrays.
                append_mention_graph_row(source_node)
                append_mention_graph_col(mention_target_node)

            # This is the first time we deal with this tweet.
            if original_tweet_id not in tweet_id_set:
                # Update sets, lists and dictionaries.
                add_tweet_id(original_tweet_id)
                id_to_name[original_tweet_user_id] = original_tweet_user_name
                append_user_id(original_tweet_user_id)

                ########################################################################################################
                # Update mention matrix.
                ########################################################################################################
                # Update the mention graph one-by-one.
                for mentioned_user_id in mentioned_user_id_set:
                    # Map users to distinct integer numbers.
                    graph_size = len(id_to_node)
                    mention_target_node = id_to_node.setdefault(mentioned_user_id, graph_size)

                    append_user_id(mentioned_user_id)

                    # Add values to the sparse matrix arrays.
                    append_mention_graph_row(original_tweet_node)
                    append_mention_graph_col(mention_target_node)
            else:
                pass

    ####################################################################################################################
    # Final steps of preprocessing tweets.
    ####################################################################################################################
    # Discard any duplicates.
    user_id_set = set(user_id_set)
    number_of_users = len(user_id_set)
    # min_number_of_users = max(user_id_set) + 1

    # Form mention graph adjacency matrix.
    mention_graph_row = np.array(mention_graph_row, dtype=np.int64)
    mention_graph_col = np.array(mention_graph_col, dtype=np.int64)
    mention_graph_data = np.ones_like(mention_graph_row, dtype=np.float64)

    mention_graph = spsp.coo_matrix((mention_graph_data, (mention_graph_row, mention_graph_col)),
                                    shape=(number_of_users, number_of_users))
    mention_graph = spsp.coo_matrix(spsp.csr_matrix(mention_graph))

    # Form retweet graph adjacency matrix.
    retweet_graph_row = np.array(retweet_graph_row, dtype=np.int64)
    retweet_graph_col = np.array(retweet_graph_col, dtype=np.int64)
    retweet_graph_data = np.ones_like(retweet_graph_row, dtype=np.float64)

    retweet_graph = spsp.coo_matrix((retweet_graph_data, (retweet_graph_row, retweet_graph_col)),
                                    shape=(number_of_users, number_of_users))
    retweet_graph = spsp.coo_matrix(spsp.csr_matrix(retweet_graph))

    node_to_id = dict(zip(id_to_node.values(), id_to_node.keys()))

    return mention_graph, retweet_graph, tweet_id_set, user_id_set, node_to_id, id_to_name


def extract_connected_components(graph, connectivity_type, node_to_id):
    """
    Extract the largest connected component from a graph.

    Inputs:  - graph: An adjacency matrix in scipy sparse matrix format.
             - connectivity_type: A string that can be either: "strong" or "weak".
             - node_to_id: A python dictionary that maps from node anonymized ids, to twitter user ids.

    Outputs: - largest_connected_component: An adjacency matrix in scipy sparse matrix format.
             - node_to_id: A map from graph node id to Twitter id, in python dictionary format.

    Raises:  - RuntimeError: If there the input graph is empty.
    """
    # Get a networkx graph.
    nx_graph = nx.from_scipy_sparse_matrix(graph, create_using=nx.DiGraph())

    # Calculate all connected components in graph.
    if connectivity_type == "weak":
        largest_connected_component_list = nxalgcom.weakly_connected_component_subgraphs(nx_graph)
    elif connectivity_type == "strong":
        largest_connected_component_list = nxalgcom.strongly_connected_component_subgraphs(nx_graph)
    else:
        print("Invalid connectivity type input.")
        raise RuntimeError

    # Handle empty graph.
    try:
        largest_connected_component = max(largest_connected_component_list, key=len)
    except ValueError:
        print("Error: Empty graph.")
        raise RuntimeError

    old_node_list = largest_connected_component.nodes()
    node_to_node = dict(zip(np.arange(len(old_node_list)), old_node_list))
    largest_connected_component = nx.to_scipy_sparse_matrix(largest_connected_component, dtype=np.float64, format="csr")

    # Make node_to_id.
    new_node_to_id = {k: node_to_id[v] for k, v in node_to_node.items()}

    return largest_connected_component, new_node_to_id, old_node_list



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import sys
import subprocess
import urllib
from amqp import Connection, Message


if sys.version_info > (3,):
    import urllib.parse as urlparse
else:
    import urlparse


def translate_rabbitmq_url(url):
    if url[0:4] == "amqp":
        url = "http" + url[4:]

    parts = urlparse.urlparse(url)

    if parts.scheme == "https":
        ssl = True
    else:
        ssl = False

    if parts.hostname is not None:
        host_name = parts.hostname
    else:
        host_name = "localhost"

    if parts.port is not None:
        port = int(parts.port)
    else:
        port = 5672

    if parts.username is not None:
        user_name = parts.username
        password = parts.password
    else:
        user_name = parts.username
        password = parts.password

    path_parts = parts.path.split('/')
    virtual_host = urllib.parse.unquote(path_parts[1])
    virtual_host = virtual_host  # May be an empty path if URL is e.g. "amqp://guest:guest@localhost:5672/vhost"
    if virtual_host == "":
        virtual_host = "/"  # Default vhost

    return user_name, password, host_name, port, virtual_host, ssl


def establish_rabbitmq_connection(rabbitmq_uri):
    """
    What it says on the tin.

    Input:  - rabbitmq_uri: A RabbitMQ URI.

    Output: - connection: A RabbitMQ connection.
    """
    userid, password, host, port, virtual_host, ssl = translate_rabbitmq_url(rabbitmq_uri)

    connection = Connection(userid=userid,
                            password=password,
                            host=host,
                            port=port,
                            virtual_host=virtual_host,
                            ssl=False)

    return connection


def simple_notification(connection, queue_name, exchange_name, routing_key, text_body):
    """
    Publishes a simple notification.

    Inputs: - connection: A rabbitmq connection object.
            - queue_name: The name of the queue to be checked or created.
            - exchange_name: The name of the notification exchange.
            - routing_key: The routing key for the exchange-queue binding.
            - text_body: The text to be published.
    """
    channel = connection.channel()
    channel.queue_declare(queue_name, durable=True)
    channel.exchange_declare(exchange_name, type='direct', durable=True)
    channel.queue_bind(queue_name, exchange_name, routing_key=routing_key)

    message = Message(text_body)
    channel.basic_publish(message, exchange_name, routing_key)


def simpler_notification(channel, queue_name, exchange_name, routing_key, text_body):
    message = Message(text_body)
    channel.basic_publish(message, exchange_name, routing_key)


def rabbitmq_server_service(command):
    subprocess.call(["service", "rabbitmq-server", command])



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import argparse
import os
from multiprocessing import Pool
from functools import partial
import json

from reveal_user_annotation.common.config_package import get_threads_number
from reveal_user_annotation.common.datarw import load_pickle
from reveal_user_annotation.text.map_data import chunks
from reveal_user_annotation.twitter.clean_twitter_list import user_twitter_list_bag_of_words


def worker_function(file_name_list,
                    lemmatizing,
                    source_folder,
                    target_folder):
    source_path_list = (source_folder + "/" + file_name for file_name in file_name_list)
    target_path_list = (target_folder + "/" + file_name[:-4] + ".json" for file_name in file_name_list)

    # Get the lists of a user
    for source_path in source_path_list:
        twitter_lists_corpus = load_pickle(source_path)
        if "lists" in twitter_lists_corpus.keys():
            twitter_lists_corpus = twitter_lists_corpus["lists"]
        else:
            continue

        bag_of_lemmas, lemma_to_keywordbag = user_twitter_list_bag_of_words(twitter_lists_corpus, lemmatizing)

        user_annotation = dict()
        user_annotation["bag_of_lemmas"] = bag_of_lemmas
        user_annotation["lemma_to_keywordbag"] = lemma_to_keywordbag

        target_path = next(target_path_list)
        with open(target_path, "w", encoding="utf-8") as fp:
            json.dump(user_annotation, fp)


def main():
    # Parse arguments.
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--source", dest="source_folder",
                        help="This is the folder with the pickled Twitter lists.",
                        type=str, required=True)
    parser.add_argument("-t", "--target", dest="target_folder",
                        help="This is the folder where the extracted keyword jsons will be stored.",
                        type=str, required=True)

    args = parser.parse_args()

    source_folder = args.source_folder
    target_folder = args.target_folder

    # Get the file names where the twitter lists for certain users are stored.
    file_name_list = os.listdir(source_folder)

    # Build a pool of processes.
    pool = Pool(processes=get_threads_number()*2,)

    # Partition dataset in chunks.
    user_chunks = chunks(file_name_list, get_threads_number()*2)

    # Extract bags of words in parallel and serialize and store in JSON format.
    pool.map(partial(worker_function,
                     lemmatizing="wordnet",
                     source_folder=source_folder,
                     target_folder=target_folder),
             user_chunks)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import requests
import xml.etree.cElementTree as etree
from io import StringIO


def get_user_list(host_name, client_name, client_pass):
    """
    Pulls the list of users in a client.

    Inputs: - host_name: A string containing the address of the machine where the PServer instance is hosted.
            - client_name: The PServer client name.
            - client_pass: The PServer client's password.

    Output: - user_id_list: A python list of user ids.

    Raises: -
    """
    # Construct request.
    request = construct_request(model_type="pers",
                                client_name=client_name,
                                client_pass=client_pass,
                                command="getusrs",
                                values="whr=*")

    # Make request.
    request_result = send_request(host_name, request)

    # Extract a python list from xml object.
    user_id_list = list()
    append_user_id = user_id_list.append

    if request_result is not None:
        user_list_xml = request_result.text
        tree = etree.parse(StringIO(user_list_xml))
        root = tree.getroot()

        xml_rows = root.findall("./result/row/usr")
        for xml_row in xml_rows:
            append_user_id(xml_row.text)

    return user_id_list


def add_features(host_name, client_name, client_pass, feature_names):
    """
    Add a number of numerical features in the client.

    Inputs: - host_name: A string containing the address of the machine where the PServer instance is hosted.
            - client_name: The PServer client name.
            - client_pass: The PServer client's password.
            - feature_names: A python list of feature names.
    """
    init_feats = ("&".join(["%s=0"]*len(feature_names))) % tuple(feature_names)
    features_req = construct_request("pers",
                                     client_name,
                                     client_pass,
                                     "addftr",
                                     init_feats)
    send_request(host_name,
                 features_req)


def delete_features(host_name, client_name, client_pass, feature_names=None):
    """
    Remove a number of numerical features in the client. If a list  is not provided, remove all features.

    Inputs: - host_name: A string containing the address of the machine where the PServer instance is hosted.
            - client_name: The PServer client name.
            - client_pass: The PServer client's password.
            - feature_names: A python list of feature names.
    """
    # Get all features.
    if feature_names is None:
        feature_names = get_feature_names(host_name,
                                          client_name,
                                          client_pass)

    # Remove all features.
    feature_to_be_removed = ("&".join(["ftr=%s"]*len(feature_names))) % tuple(feature_names)
    features_req = construct_request("pers",
                                     client_name,
                                     client_pass,
                                     'remftr',
                                     feature_to_be_removed)
    send_request(host_name,
                 features_req)


def get_feature_names(host_name, client_name, client_pass):
    """
    Get the names of all features in a PServer client.

    Inputs: - host_name: A string containing the address of the machine where the PServer instance is hosted.
            - client_name: The PServer client name.
            - client_pass: The PServer client's password.

    Output: - feature_names: A python list of feature names.
    """
    # Construct request.
    request = construct_request(model_type="pers",
                                client_name=client_name,
                                client_pass=client_pass,
                                command="getftrdef",
                                values="ftr=*")

    # Send request.
    request_result = send_request(host_name,
                                  request)

    # Extract a python list from xml object.
    feature_names = list()
    append_feature_name = feature_names.append

    if request_result is not None:
        feature_names_xml = request_result.text
        tree = etree.parse(StringIO(feature_names_xml))
        root = tree.getroot()

        xml_rows = root.findall("row/ftr")
        for xml_row in xml_rows:
            append_feature_name(xml_row.text)

    return feature_names


def insert_user_data(host_name, client_name, client_pass, user_twitter_id, topic_to_score):
    """
    Inserts topic/score data for a user to a PServer client.

    Inputs: - host_name: A string containing the address of the machine where the PServer instance is hosted.
            - client_name: The PServer client name.
            - client_pass: The PServer client's password.
            - feature_names: A python list of feature names.
            - user_twitter_id: A Twitter user identifier.
            - topic_to_score: A python dictionary that maps from topic to score.
    """

    # Construct values.
    values = "usr=" + str(user_twitter_id)

    for topic, score in topic_to_score.items():
        values += "&type." + topic + "=%.2f" % score

    # Construct request.
    request = construct_request(model_type="pers",
                                client_name=client_name,
                                client_pass=client_pass,
                                command="setusr",
                                values=values)

    # Send request.
    send_request(host_name,
                 request)


def construct_request(model_type, client_name, client_pass, command, values):
    """
    Construct the request url.

    Inputs: - model_type: PServer usage mode type.
            - client_name: The PServer client name.
            - client_pass: The PServer client's password.
            - command: A PServer command.
            - values: PServer command arguments.

    Output: - base_request: The base request string.
    """
    base_request = ("{model_type}?"
                    "clnt={client_name}|{client_pass}&"
                    "com={command}&{values}".format(model_type=model_type,
                                                    client_name=client_name,
                                                    client_pass=client_pass,
                                                    command=command,
                                                    values=values))
    return base_request


def send_request(host_name, request):
    """
    Sends a PServer url request.

    Inputs: - host_name: A string containing the address of the machine where the PServer instance is hosted.
            - request: The url request.
    """
    request = "%s%s" % (host_name, request)
    # print(request)
    try:
        result = requests.get(request)
        if result.status_code == 200:
            return result
        else:
            # print(result.status_code)
            raise Exception
    except Exception as e:
        # print(e)
        raise e


def update_feature_value(host_name, client_name, client_pass, user_twitter_id, feature_name, feature_score):
    """
    Updates a single topic score, for a single user.
    """
    username = str(user_twitter_id)
    feature_value = "{0:.2f}".format(feature_score)
    joined_ftr_value = "ftr_" + feature_name + "=" + str(feature_value)
    values = "usr=%s&%s" % (username, joined_ftr_value)

    # Construct request.
    request = construct_request(model_type="pers",
                                client_name=client_name,
                                client_pass=client_pass,
                                command="setusr",
                                values=values)

    # Send request.
    send_request(host_name,
                 request)



# -*- coding: <UTF-8> -*-
__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import time
import numpy as np
import scipy.sparse as sparse
import copy
import twython
from collections import defaultdict
from operator import itemgetter
from urllib.error import URLError
from http.client import BadStatusLine

from reveal_user_annotation.text.text_util import augmented_tf_idf, simple_word_query
from reveal_user_annotation.text.clean_text import clean_single_word
from reveal_user_annotation.twitter.twitter_util import login, safe_twitter_request_handler
from reveal_user_annotation.twitter.clean_twitter_list import user_twitter_list_bag_of_words
from reveal_user_annotation.twitter.manage_resources import get_reveal_set, get_topic_keyword_dictionary


def extract_user_keywords_generator(twitter_lists_gen, lemmatizing="wordnet"):
    """
    Based on the user-related lists I have downloaded, annotate the users.

    Inputs: - twitter_lists_gen: A python generator that yields a user Twitter id and a generator of Twitter lists.
            - lemmatizing: A string containing one of the following: "porter", "snowball" or "wordnet".

    Yields: - user_twitter_id: A Twitter user id.
            - user_annotation: A python dictionary that contains two dicts:
                * bag_of_lemmas: Maps emmas to multiplicity.
                * lemma_to_keywordbag: A python dictionary that maps stems/lemmas to original topic keywords.
    """
    ####################################################################################################################
    # Extract keywords serially.
    ####################################################################################################################
    for user_twitter_id, twitter_lists_list in twitter_lists_gen:
        if twitter_lists_list is not None:
            if "lists" in twitter_lists_list.keys():
                twitter_lists_list = twitter_lists_list["lists"]

            bag_of_lemmas, lemma_to_keywordbag = user_twitter_list_bag_of_words(twitter_lists_list, lemmatizing)

            for lemma, keywordbag in lemma_to_keywordbag.items():
                lemma_to_keywordbag[lemma] = dict(keywordbag)
            lemma_to_keywordbag = dict(lemma_to_keywordbag)

            user_annotation = dict()
            user_annotation["bag_of_lemmas"] = bag_of_lemmas
            user_annotation["lemma_to_keywordbag"] = lemma_to_keywordbag

            yield user_twitter_id, user_annotation


def form_user_label_matrix(user_twitter_list_keywords_gen, id_to_node, max_number_of_labels):
    """
    Forms the user-label matrix to be used in multi-label classification.

    Input:   - user_twitter_list_keywords_gen:
             - id_to_node:  A Twitter id to node map as a python dictionary.

    Outputs: - user_label_matrix: A user-to-label matrix in scipy sparse matrix format.
             - annotated_nodes: A numpy array containing graph nodes.
             - label_to_lemma: A python dictionary that maps a numerical label to a string topic lemma.
             - lemma_to_keyword: A python dictionary that maps a lemma to the original keyword.
    """
    user_label_matrix, annotated_nodes, label_to_lemma, node_to_lemma_tokeywordbag = form_user_term_matrix(user_twitter_list_keywords_gen,
                                                                                                           id_to_node,
                                                                                                           None)

    # write_terms_and_frequencies("/home/georgerizos/Documents/term_matrix.txt", user_label_matrix, label_to_lemma)

    user_label_matrix, annotated_nodes, label_to_lemma = filter_user_term_matrix(user_label_matrix,
                                                                                 annotated_nodes,
                                                                                 label_to_lemma,
                                                                                 max_number_of_labels)

    # write_terms_and_frequencies("/home/georgerizos/Documents/label_matrix.txt", user_label_matrix, label_to_lemma)

    lemma_to_keyword = form_lemma_tokeyword_map(annotated_nodes, node_to_lemma_tokeywordbag)

    return user_label_matrix, annotated_nodes, label_to_lemma, lemma_to_keyword


def write_terms_and_frequencies(path, user_term_matrix, label_to_lemma):
    from collections import OrderedDict

    bag_of_words = defaultdict(int)

    matrix = sparse.coo_matrix(user_term_matrix)

    for e in range(matrix.getnnz()):
        term = matrix.col[e]
        lemma = label_to_lemma[term]
        multiplicity = matrix.data[e]

        bag_of_words[lemma] += multiplicity

    bag_of_words = dict(bag_of_words)
    bag_of_words = OrderedDict(sorted(bag_of_words.items(), key=lambda t: t[1]))

    with open(path, "w") as fp:
        for key, value in bag_of_words.items():
            row = str(key) + "\t" + str(value) + "\n"
            fp.write(row)


def form_user_term_matrix(user_twitter_list_keywords_gen, id_to_node, lemma_set=None, keyword_to_topic_manual=None):
    """
    Forms a user-term matrix.

    Input:   - user_twitter_list_keywords_gen: A python generator that yields a user Twitter id and a bag-of-words.
             - id_to_node:  A Twitter id to node map as a python dictionary.
             - lemma_set: For the labelling, we use only lemmas in this set. Default: None

    Outputs: - user_term_matrix: A user-to-term matrix in scipy sparse matrix format.
             - annotated_nodes: A numpy array containing graph nodes.
             - label_to_topic: A python dictionary that maps a numerical label to a string topic/keyword.
             - node_to_lemma_tokeywordbag: A python dictionary that maps nodes to lemma-to-keyword bags.
    """
    # Prepare for iteration.
    term_to_attribute = dict()

    user_term_matrix_row = list()
    user_term_matrix_col = list()
    user_term_matrix_data = list()

    append_user_term_matrix_row = user_term_matrix_row.append
    append_user_term_matrix_col = user_term_matrix_col.append
    append_user_term_matrix_data = user_term_matrix_data.append

    annotated_nodes = list()
    append_node = annotated_nodes.append

    node_to_lemma_tokeywordbag = dict()

    invalid_terms = list()
    counter = 0

    if keyword_to_topic_manual is not None:
        manual_keyword_list = list(keyword_to_topic_manual.keys())

    for user_twitter_id, user_annotation in user_twitter_list_keywords_gen:
        counter += 1
        # print(counter)
        bag_of_words = user_annotation["bag_of_lemmas"]
        lemma_to_keywordbag = user_annotation["lemma_to_keywordbag"]

        if lemma_set is not None:
            bag_of_words = {lemma: multiplicity for lemma, multiplicity in bag_of_words.items() if lemma in lemma_set}
            lemma_to_keywordbag = {lemma: keywordbag for lemma, keywordbag in lemma_to_keywordbag.items() if lemma in lemma_set}

        node = id_to_node[user_twitter_id]
        append_node(node)
        node_to_lemma_tokeywordbag[node] = lemma_to_keywordbag
        for term, multiplicity in bag_of_words.items():
            if term == "news":
                continue

            if keyword_to_topic_manual is not None:
                keyword_bag = lemma_to_keywordbag[term]
                term = max(keyword_bag.keys(), key=(lambda key: keyword_bag[key]))

                found_list_of_words = simple_word_query(term, manual_keyword_list, edit_distance=1)

                if len(found_list_of_words) > 0:
                    term = found_list_of_words[0]

                try:
                    term = keyword_to_topic_manual[term]
                except KeyError:
                    print(term)

            vocabulary_size = len(term_to_attribute)
            attribute = term_to_attribute.setdefault(term, vocabulary_size)

            append_user_term_matrix_row(node)
            append_user_term_matrix_col(attribute)
            append_user_term_matrix_data(multiplicity)

    annotated_nodes = np.array(list(set(annotated_nodes)), dtype=np.int64)

    user_term_matrix_row = np.array(user_term_matrix_row, dtype=np.int64)
    user_term_matrix_col = np.array(user_term_matrix_col, dtype=np.int64)
    user_term_matrix_data = np.array(user_term_matrix_data, dtype=np.float64)

    user_term_matrix = sparse.coo_matrix((user_term_matrix_data,
                                          (user_term_matrix_row, user_term_matrix_col)),
                                         shape=(len(id_to_node), len(term_to_attribute)))

    label_to_topic = dict(zip(term_to_attribute.values(), term_to_attribute.keys()))

    # print(user_term_matrix.shape)
    # print(len(label_to_topic))
    # print(invalid_terms)

    return user_term_matrix, annotated_nodes, label_to_topic, node_to_lemma_tokeywordbag


def filter_user_term_matrix(user_term_matrix, annotated_nodes, label_to_topic, max_number_of_labels=None):
    """
    Filters out labels that are either too rare, or have very few representatives.

    Inputs:  - user_term_matrix: A user-to-term matrix in scipy sparse matrix format.
             - annotated_nodes: A numpy array containing graph nodes.
             - label_to_topic: A python dictionary that maps a numerical label to a string topic/keyword.
             - max_number_of_labels: The upper limit for the number of labels.

    Outputs: - user_term_matrix: The filtered user-to-term matrix in scipy sparse matrix format.
             - annotated_nodes: A numpy array containing graph nodes.
             - label_to_topic: A python dictionary that maps a numerical label to a string topic/keyword.
    """
    # Aggregate edges and eliminate zeros.
    user_term_matrix = user_term_matrix.tocsr()
    user_term_matrix.eliminate_zeros()

    ####################################################################################################################
    # Enforce max number of labels if required.
    ####################################################################################################################
    # Form matrix that indicates which user-label pairs are nonzero.
    temp_matrix = copy.copy(user_term_matrix)
    temp_matrix.data = np.ones_like(temp_matrix.data).astype(np.int64)
    temp_matrix.astype(np.int64)

    # Calculate the label-user counts.
    label_frequencies = temp_matrix.sum(axis=0)
    index = np.argsort(np.squeeze(np.asarray(label_frequencies)))

    # Apply max number of labels cutoff.
    if max_number_of_labels is not None:
        if index.size > max_number_of_labels:
            index = index[:index.size-max_number_of_labels]
    else:
        index = np.array(list())

    # If there are zero samples of a label, remove them from both the matrix and the label-to-topic map.

    # percentile = 90
    # p = np.percentile(label_frequencies, percentile)
    # index = np.where(label_frequencies <= p)[1]
    if index.size > 0:
        index = np.squeeze(np.asarray(index))
        index = np.setdiff1d(np.arange(label_frequencies.size), index)
        user_term_matrix = user_term_matrix[:, index]
        temp_map = dict()
        counter = 0
        for i in index:
            temp_map[counter] = label_to_topic[i]
            counter += 1
        label_to_topic = temp_map

    # print(user_term_matrix.getnnz())
    # print(user_term_matrix.shape)

    # Perform tf-idf on all matrices.
    user_term_matrix = augmented_tf_idf(user_term_matrix)

    ####################################################################################################################
    # Most topics are heavy-tailed. For each topic, threshold to annotate.
    ####################################################################################################################
    user_term_nnz = user_term_matrix.getnnz()
    # percentile = 80
    percentile = 50

    while True:
        matrix_row = list()
        matrix_col = list()
        extend_matrix_row = matrix_row.extend
        extend_matrix_col = matrix_col.extend

        user_term_matrix = sparse.csc_matrix(user_term_matrix)
        for topic in range(user_term_matrix.shape[1]):
            col = user_term_matrix.getcol(topic)
            p = np.percentile(col.data, percentile)
            index = col.indices[col.data >= p]
            extend_matrix_row(index)
            extend_matrix_col(topic*np.ones_like(index))
        matrix_row = np.array(matrix_row, dtype=np.int64)
        matrix_col = np.array(matrix_col, dtype=np.int64)
        matrix_data = np.ones_like(matrix_row, dtype=np.int8)
        user_term_matrix = sparse.coo_matrix((matrix_data, (matrix_row, matrix_col)), shape=user_term_matrix.shape)
        user_term_matrix = sparse.csr_matrix(user_term_matrix)

        if (user_term_matrix.getnnz() > user_term_nnz/10) and (user_term_matrix.shape[1] > 1):
            break
        else:
            percentile -= 10
            if percentile <= 10:
                break

    # print(user_term_matrix.getnnz())
    # print(user_term_matrix.shape)

    ####################################################################################################################
    # If there is a small number of samples of a label, remove them from both the matrix and the label-to-topic map.
    ####################################################################################################################
    user_term_nnz = user_term_matrix.getnnz()

    minimum_frequency = 3

    while True:
        # Form matrix that indicates which user-label pairs are nonzero.
        temp_matrix = copy.copy(user_term_matrix)
        temp_matrix.data = np.ones_like(temp_matrix.data).astype(np.int64)
        temp_matrix.astype(np.int64)

        label_frequencies = temp_matrix.sum(axis=0)

        # index = np.where(label_frequencies < 31)[1]  # Fewer than 10.
        index = np.where(label_frequencies < minimum_frequency)[1]  # Fewer than 10.
        if index.shape[1] > 0:
            index = np.squeeze(np.asarray(index))
            index = np.setdiff1d(np.arange(label_frequencies.size), index)
            user_term_matrix = user_term_matrix[:, index]
            temp_map = dict()
            counter = 0
            for i in index:
                temp_map[counter] = label_to_topic[i]
                counter += 1
            label_to_topic = temp_map

        if (user_term_matrix.getnnz() > user_term_nnz/2) and (user_term_matrix.shape[1] > 1):
            break
        else:
            minimum_frequency -= 1
            if minimum_frequency <= 1:
                break

    # print(user_term_matrix.getnnz())
    # print(user_term_matrix.shape)

    # print(len(label_to_topic))

    return user_term_matrix, annotated_nodes, label_to_topic


# def filter_user_term_matrix(user_term_matrix, annotated_nodes, label_to_topic, max_number_of_labels=None):
#     """
#     Filters out labels that are either too rare, or have very few representatives.
#
#     Inputs:  - user_term_matrix: A user-to-term matrix in scipy sparse matrix format.
#              - annotated_nodes: A numpy array containing graph nodes.
#              - label_to_topic: A python dictionary that maps a numerical label to a string topic/keyword.
#              - max_number_of_labels: The upper limit for the number of labels.
#
#     Outputs: - user_term_matrix: The filtered user-to-term matrix in scipy sparse matrix format.
#              - annotated_nodes: A numpy array containing graph nodes.
#              - label_to_topic: A python dictionary that maps a numerical label to a string topic/keyword.
#     """
#     # Aggregate edges and eliminate zeros.
#     user_term_matrix = user_term_matrix.tocsr()
#     user_term_matrix.eliminate_zeros()
#
#     # Form matrix that indicates which user-label pairs are nonzero.
#     temp_matrix = copy.copy(user_term_matrix)
#     temp_matrix.data = np.ones_like(temp_matrix.data).astype(np.int64)
#     # temp_matrix.astype(np.int64)
#
#     # Calculate the label-user counts.
#     label_frequencies = temp_matrix.sum(axis=0)
#     index = np.argsort(np.squeeze(np.asarray(label_frequencies)))
#
#     # Apply max number of labels cutoff.
#     if max_number_of_labels is not None:
#         if index.size > max_number_of_labels:
#             index = index[:index.size-max_number_of_labels]
#     else:
#         index = np.array(list())
#
#     # If there are zero samples of a label, remove them from both the matrix and the label-to-topic map.
#
#     # percentile = 90
#     # p = np.percentile(label_frequencies, percentile)
#     # index = np.where(label_frequencies <= p)[1]
#     if index.size > 0:
#         index = np.squeeze(np.asarray(index))
#         index = np.setdiff1d(np.arange(label_frequencies.size), index)
#         user_term_matrix = user_term_matrix[:, index]
#         temp_map = dict()
#         counter = 0
#         for i in index:
#             temp_map[counter] = label_to_topic[i]
#             counter += 1
#         label_to_topic = temp_map
#
#     # Perform tf-idf on all matrices.
#     user_term_matrix = augmented_tf_idf(user_term_matrix)
#
#     # Most topics are heavy-tailed. For each topic, threshold to annotate.
#     percentile = 80
#
#     matrix_row = list()
#     matrix_col = list()
#     extend_matrix_row = matrix_row.extend
#     extend_matrix_col = matrix_col.extend
#
#     user_term_matrix = sparse.csc_matrix(user_term_matrix)
#     for topic in range(user_term_matrix.shape[1]):
#         col = user_term_matrix.getcol(topic)
#         p = np.percentile(col.data, percentile)
#         index = col.indices[col.data >= p]
#         extend_matrix_row(index)
#         extend_matrix_col(topic*np.ones_like(index))
#     matrix_row = np.array(matrix_row, dtype=np.int64)
#     matrix_col = np.array(matrix_col, dtype=np.int64)
#     matrix_data = np.ones_like(matrix_row, dtype=np.int8)
#     user_term_matrix = sparse.coo_matrix((matrix_data, (matrix_row, matrix_col)), shape=user_term_matrix.shape)
#     user_term_matrix = sparse.csr_matrix(user_term_matrix)
#
#     # Form matrix that indicates which user-label pairs are nonzero.
#     temp_matrix = copy.copy(user_term_matrix)
#     temp_matrix.data = np.ones_like(temp_matrix.data).astype(np.int64)
#     # temp_matrix.astype(np.int64)
#
#     # If there is a small number of samples of a label, remove them from both the matrix and the label-to-topic map.
#     label_frequencies = temp_matrix.sum(axis=0)
#     index = np.where(label_frequencies < 31)[1]  # Fewer than 10.
#     if index.shape[1] > 0:
#         index = np.squeeze(np.asarray(index))
#         index = np.setdiff1d(np.arange(label_frequencies.size), index)
#         user_term_matrix = user_term_matrix[:, index]
#         temp_map = dict()
#         counter = 0
#         for i in index:
#             temp_map[counter] = label_to_topic[i]
#             counter += 1
#         label_to_topic = temp_map
#
#     # print(user_term_matrix.shape)
#     # print(len(label_to_topic))
#
#     return user_term_matrix, annotated_nodes, label_to_topic


def semi_automatic_user_annotation(user_twitter_list_keywords_gen, id_to_node):

    reveal_set = get_reveal_set()
    topic_keyword_dict = get_topic_keyword_dictionary()

    available_topics = set(list(topic_keyword_dict.keys()))

    keyword_list = list()
    for topic in reveal_set:
        if topic in available_topics:
            keyword_list.extend(topic_keyword_dict[topic])

    lemma_set = list()
    for keyword in keyword_list:
        lemma = clean_single_word(keyword, lemmatizing="wordnet")
        lemma_set.append(lemma)
    lemma_set = set(lemma_set)

    keyword_topic_dict = dict()
    for topic, keyword_set in topic_keyword_dict.items():
        for keyword in keyword_set:
            keyword_topic_dict[keyword] = topic

    user_label_matrix, annotated_nodes, label_to_lemma, node_to_lemma_tokeywordbag = form_user_term_matrix(user_twitter_list_keywords_gen, id_to_node, lemma_set=lemma_set, keyword_to_topic_manual=keyword_topic_dict)


    user_label_matrix, annotated_nodes, label_to_lemma = filter_user_term_matrix(user_label_matrix,
                                                                                 annotated_nodes,
                                                                                 label_to_lemma,
                                                                                 max_number_of_labels=None)

    lemma_to_keyword = form_lemma_tokeyword_map(annotated_nodes, node_to_lemma_tokeywordbag)

    return user_label_matrix, annotated_nodes, label_to_lemma, lemma_to_keyword


def form_lemma_tokeyword_map(annotated_nodes, node_to_lemma_tokeywordbag):
    """
    Forms the aggregated dictionary that maps lemmas/stems to the most popular topic keyword.

    Inputs: - annotated_nodes: A numpy array of anonymized nodes that are annotated.
            - node_to_lemma_tokeywordbag: A map from nodes to maps from lemmas to bags of keywords.

    Output: - lemma_to_keyword: A dictionary that maps lemmas to keywords.
    """
    # Reduce relevant lemma-to-original keyword bags.
    lemma_to_keywordbag = defaultdict(lambda: defaultdict(int))
    for node in annotated_nodes:
        for lemma, keywordbag in node_to_lemma_tokeywordbag[node].items():
            for keyword, multiplicity in keywordbag.items():
                lemma_to_keywordbag[lemma][keyword] += multiplicity

    lemma_to_keyword = dict()
    for lemma, keywordbag in lemma_to_keywordbag.items():
        lemma_to_keyword[lemma] = max(keywordbag.items(), key=itemgetter(1))[0]

    return lemma_to_keyword


def fetch_twitter_lists_for_user_ids_generator(twitter_app_key,
                                               twitter_app_secret,
                                               user_id_list):
    """
    Collects at most 500 Twitter lists for each user from an input list of Twitter user ids.

    Inputs: - twitter_app_key: What is says on the tin.
            - twitter_app_secret: Ditto.
            - user_id_list: A python list of Twitter user ids.

    Yields: - user_twitter_id: A Twitter user id.
            - twitter_lists_list: A python list containing Twitter lists in dictionary (json) format.
    """
    ####################################################################################################################
    # Log into my application.
    ####################################################################################################################
    twitter = login(twitter_app_key,
                    twitter_app_secret)

    ####################################################################################################################
    # For each user, gather at most 500 Twitter lists.
    ####################################################################################################################
    get_list_memberships_counter = 0
    get_list_memberships_time_window_start = time.perf_counter()
    for user_twitter_id in user_id_list:
        # Make safe twitter request.
        try:
            twitter_lists_list, get_list_memberships_counter, get_list_memberships_time_window_start\
                = safe_twitter_request_handler(twitter_api_func=twitter.get_list_memberships,
                                               call_rate_limit=15,
                                               call_counter=get_list_memberships_counter,
                                               time_window_start=get_list_memberships_time_window_start,
                                               max_retries=5,
                                               wait_period=2,
                                               user_id=user_twitter_id,
                                               count=500,
                                               cursor=-1)
            # If the call is succesful, yield the list of Twitter lists.
            yield user_twitter_id, twitter_lists_list
        except twython.TwythonError:
            # If the call is unsuccesful, we do not have any Twitter lists to store.
            yield user_twitter_id, None
        except URLError:
            # If the call is unsuccesful, we do not have any Twitter lists to store.
            yield user_twitter_id, None
        except BadStatusLine:
            # If the call is unsuccesful, we do not have any Twitter lists to store.
            yield user_twitter_id, None


def decide_which_users_to_annotate(centrality_vector,
                                   number_to_annotate,
                                   already_annotated,
                                   node_to_id):
    """
    Sorts a centrality vector and returns the Twitter user ids that are to be annotated.

    Inputs: - centrality_vector: A numpy array vector, that contains the centrality values for all users.
            - number_to_annotate: The number of users to annotate.
            - already_annotated: A python set of user twitter ids that have already been annotated.
            - node_to_id: A python dictionary that maps graph nodes to user twitter ids.

    Output: - user_id_list: A python list of Twitter user ids.
    """
    # Sort the centrality vector according to decreasing centrality.
    centrality_vector = np.asarray(centrality_vector)
    ind = np.argsort(np.squeeze(centrality_vector))
    if centrality_vector.size > 1:
        reversed_ind = ind[::-1]
    else:
        reversed_ind = list()
        reversed_ind = reversed_ind.append(ind)

    # Get the sublist of Twitter user ids to return.
    user_id_list = list()
    append_user_id = user_id_list.append

    counter = 0
    for node in reversed_ind:
        user_twitter_id = node_to_id[node]
        if user_twitter_id not in already_annotated:
            append_user_id(user_twitter_id)
            counter += 1

            if counter >= number_to_annotate:
                break

    return user_id_list


def on_demand_annotation(twitter_app_key, twitter_app_secret, user_twitter_id):
    """
    A service that leverages twitter lists for on-demand annotation of popular users.

    TODO: Do this.
    """
    ####################################################################################################################
    # Log into my application
    ####################################################################################################################
    twitter = login(twitter_app_key, twitter_app_secret)

    twitter_lists_list = twitter.get_list_memberships(user_id=user_twitter_id, count=1000)

    for twitter_list in twitter_lists_list:
        print(twitter_list)

    return twitter_lists_list



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from setuptools import setup


def readme():
    with open("README.md") as f:
        return f.read()

setup(
    name='reveal-user-annotation',
    version='0.2.2',
    author='Georgios Rizos',
    author_email='georgerizos@iti.gr',
    packages=['reveal_user_annotation',
              'reveal_user_annotation.common',
              'reveal_user_annotation.text',
              'reveal_user_annotation.twitter',
              'reveal_user_annotation.mongo',
              'reveal_user_annotation.pserver',
              'reveal_user_annotation.rabbitmq',
              'reveal_user_annotation.entry_points'],
    url='https://github.com/MKLab-ITI/reveal-user-annotation',
    license='Apache',
    description='Utility methods for generating labels for Twitter users and handling their storage and retrieval',
    long_description=readme(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],
    keywords="online-social-network user-annotation twitter-list-crowdsourcing Reveal-FP7",
    entry_points={
        'console_scripts': ['store_snow_tweets_in_mongo=reveal_user_annotation.entry_points.store_snow_tweets_in_mongo:main',
                            'extract_twitter_list_keywords=reveal_user_annotation.entry_points.extract_twitter_list_keywords:main'],
    },
    package_data={'reveal_user_annotation.text': ['res/stopwords/*.txt'],
                  'reveal_user_annotation.twitter': ['res/topics/*.txt']},
    include_package_data=False,
    install_requires=open("requirements.txt").read().split("\n")
)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import twython
from twython import Twython
import time
from urllib.error import URLError
from http.client import BadStatusLine

from reveal_user_annotation.common.config_package import get_package_path
from reveal_user_annotation.common.datarw import get_file_row_generator


def login(twitter_app_key,
          twitter_app_secret):
    """
    This is a shortcut for logging in the Twitter app described in the config_app_credentials.txt file.

    Output: - twitter: A twython twitter object, containing wrappers for the Twitter API.
    """
    # Log into my application
    twitter = Twython(twitter_app_key, twitter_app_secret)
    # auth = twitter.get_authentication_tokens(screen_name=screen_name)

    return twitter


def safe_twitter_request_handler(twitter_api_func,
                                 call_rate_limit,
                                 call_counter,
                                 time_window_start,
                                 max_retries,
                                 wait_period,
                                 *args, **kw):
    """
    This is a safe function handler for any twitter request.

    Inputs:  - twitter_api_func: The twython function object to be safely called.
             - call_rate_limit: THe call rate limit for this specific Twitter API function.
             - call_counter: A counter that keeps track of the number of function calls in the current 15-minute window.
             - time_window_start: The timestamp of the current 15-minute window.
             - max_retries: Number of call retries allowed before abandoning the effort.
             - wait_period: For certain Twitter errors (i.e. server overload), we wait and call again.
             - *args, **kw: The parameters of the twython function to be called.

    Outputs: - twitter_api_function_result: The results of the Twitter function.
             - call_counter: A counter that keeps track of the number of function calls in the current 15-minute window.
             - time_window_start: The timestamp of the current 15-minute window.

    Raises: - twython.TwythonError
            - urllib.error.URLError
            - http.client.BadStatusLine
    """
    error_count = 0

    while True:
        try:
            # If we have reached the call rate limit for this function:
            if call_counter >= call_rate_limit:
                # Reset counter.
                call_counter = 0

                # Sleep for the appropriate time.
                elapsed_time = time.perf_counter() - time_window_start
                sleep_time = 15*60 - elapsed_time
                if sleep_time < 0.1:
                    sleep_time = 0.1
                time.sleep(sleep_time)

                # Initialize new 15-minute time window.
                time_window_start = time.perf_counter()
            else:
                call_counter += 1
                twitter_api_function_result = twitter_api_func(*args, **kw)
                return twitter_api_function_result, call_counter, time_window_start
        except twython.TwythonError as e:
            # If it is a Twitter error, handle it.
            error_count, call_counter, time_window_start, wait_period = handle_twitter_http_error(e,
                                                                                                  error_count,
                                                                                                  call_counter,
                                                                                                  time_window_start,
                                                                                                  wait_period)
            if error_count > max_retries:
                print("Max error count reached. Abandoning effort.")
                raise e
        except URLError as e:
            error_count += 1
            if error_count > max_retries:
                print("Max error count reached. Abandoning effort.")
                raise e
        except BadStatusLine as e:
            error_count += 1
            if error_count > max_retries:
                print("Max error count reached. Abandoning effort.")
                raise e


def handle_twitter_http_error(e, error_count, call_counter, time_window_start, wait_period):
    """
    This function handles the twitter request in case of an HTTP error.

    Inputs:  - e: A twython.TwythonError instance to be handled.
             - error_count: Number of failed retries of the call until now.
             - call_counter: A counter that keeps track of the number of function calls in the current 15-minute window.
             - time_window_start: The timestamp of the current 15-minute window.
             - wait_period: For certain Twitter errors (i.e. server overload), we wait and call again.

    Outputs: - call_counter: A counter that keeps track of the number of function calls in the current 15-minute window.
             - time_window_start: The timestamp of the current 15-minute window.
             - wait_period: For certain Twitter errors (i.e. server overload), we wait and call again.

    Raises: - twython.TwythonError
    """
    if e.error_code == 401:
        # Encountered 401 Error (Not Authorized)
        raise e
    elif e.error_code == 404:
        # Encountered 404 Error (Not Found)
        raise e
    elif e.error_code == 429:
        # Encountered 429 Error (Rate Limit Exceeded)
        # Sleep for 15 minutes
        error_count += 0.5
        call_counter = 0
        wait_period = 2
        time.sleep(60*15 + 5)
        time_window_start = time.perf_counter()
        return error_count, call_counter, time_window_start, wait_period
    elif e.error_code in (500, 502, 503, 504):
        error_count += 1
        time.sleep(wait_period)
        wait_period *= 1.5
        return error_count, call_counter, time_window_start, wait_period
    else:
        raise e



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import argparse

from reveal_user_annotation.mongo.store_snow_data import store_snow_tweets_from_disk_to_mongodb


def main():
    # Parse arguments.
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--folder", dest="snow_tweets_folder",
                        help="This is the folder with the SNOW tweets.",
                        type=str, required=True)

    args = parser.parse_args()

    snow_tweets_folder = args.snow_tweets_folder

    store_snow_tweets_from_disk_to_mongodb(snow_tweets_folder)
